/*
 * Radiometrix RPC radio: Kernel Device Driver
 *
 * Jeremy Elson
 * USC/ISI
 *
 * $Id: rpc-km.c,v 1.39 2001/03/12 18:46:19 jelson Exp $
 */

#define MIN(x, y) ((x) < (y) ? (x) : y)

#define __RPC_KM_C__
#define RPCPROC

#include <linux/config.h>
#include <linux/stddef.h>
#include <linux/tqueue.h>
#include <linux/sched.h>
#include <linux/kernel.h>
#ifdef MODULE
#include <linux/module.h>
#endif
#include <linux/timer.h>
#include <linux/delay.h>
#include <linux/errno.h>
#include <linux/malloc.h>
#include <linux/types.h>
#include <linux/fs.h>
#include <linux/poll.h>
#include <linux/proc_fs.h>

/* network interface headers */
#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/skbuff.h>
#include <linux/random.h>

#include <asm/uaccess.h>

#include "parapin.h"
#include "rpc_lowlevel.h"
#include "afa_fragmentation.h"
#include "krpc.h"

#ifndef MODULE
#define MOD_INC_USE_COUNT
#define MOD_DEC_USE_COUNT
#endif

/* all of these should eventually be put into a structure that has
 * one instance per RPC instance... someday... */
static int open_files = 0;
static int write_pending = 0;
static int reset_needed = 0;
static struct timer_list rpc_timer;
static struct timeval irq_time;
static struct timeval unknown_time;
static struct tq_struct rpc_queue = {0, 0, 0, 0};
static struct wait_queue *rpc_read_wait = NULL;
static struct wait_queue *rpc_write_wait = NULL;
static rpc_packet_t rpc_inq[RPC_BUFFER_SIZE];
static rpc_packet_t rpc_outq[RPC_BUFFER_SIZE];
static int rpc_inq_head = 0, rpc_inq_tail = 0;
static int rpc_outq_head = 0, rpc_outq_tail = 0;
static int rpc_mode = 0;
static int rpc_driver_on = 0;
static int rpc_holdoff_expiration = 0;

/* configurable parameters */
static int lpt = 0;
static int irq_debug = 0;
static int krpc_debug_level = 1;
static int krpc_major_number = RPC_MAJOR_NUMBER_DEFAULT;
static int ca_active = 0;
static int ca_holdoff = CA_HOLDOFF_DEFAULT;
static int ca_holdoff_rand = CA_HOLDOFF_RAND_DEFAULT;
static int rx_frames = 0;
static int rx_bytes = 0;
static int tx_frames = 0;
static int tx_bytes = 0;

#ifdef RPCPROC
/* stuff needed for /proc entries */
static char *rpc_proc_entries[MAX_RPC_PROC_ENTRIES];
static int num_rpc_proc_entries = 0;
#endif

#ifdef MODULE

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,2,0)
#error Linux 2.2 kernels are required for this driver
#endif

MODULE_PARM(lpt, "i");
MODULE_PARM(irq_debug, "i");
MODULE_PARM(krpc_debug_level, "i");
MODULE_PARM(krpc_major_number, "i");
MODULE_PARM(ca_active, "i");
MODULE_PARM(ca_holdoff, "i");
MODULE_PARM(ca_holdoff_rand, "i");

#endif

int rpc_net_init(struct device *dev);

char rpc_if_name[16];
struct device rpc_netif_dev[] = {
{
  rpc_if_name, /* name -- set at load time */
  0, 0, 0, 0,  /* shmem addresses */
  0x000,       /* ioport */
  0,           /* irq line */
  0, 0, 0,     /* various flags: init to 0 */
  NULL,        /* next ptr */
  rpc_net_init,  /* init function, fill other fields with NULL's */
}};


/************************** Boring Utilities ********************************/

#define IN_HOLDOFF_PERIOD (ca_active && (jiffies <= rpc_holdoff_expiration))
#define RPC_INQ_EMPTY (rpc_inq_head == rpc_inq_tail)
#define RPC_INQ_FULL  (rpc_inq_head == ((rpc_inq_tail+1) % RPC_BUFFER_SIZE))
#define RPC_OUTQ_EMPTY (rpc_outq_head == rpc_outq_tail)
#define RPC_OUTQ_FULL  (rpc_outq_head == ((rpc_outq_tail+1) % RPC_BUFFER_SIZE))



#define BUFSIZE 1024
asmlinkage void rdebug_real(char *fmt, ...)
{
  va_list ap;
  int len;

  /* THIS MUST BE STATIC... I HAVE NO IDEA WHY BUT IT TOOK ME 3
   * $#*&*@#$ DAYS TO TRACK DOWN THIS BUG.... something weird in the
   * kernel version of vsprintf makes it impossible to use an auto
   * (on-stack) variable for this.  no clue why. */
  static char message[BUFSIZE];

  /* THIS REALLY SHOULD BE VSNPRINTF.. pray that it's big enough */
  va_start(ap, fmt);
  len = vsprintf(message, fmt, ap);
  va_end(ap);

  if (len > BUFSIZE) {
    printk("WARNING: POSSIBLE KERNEL CORRUPTION; MESSAGE TOO LONG\n");
  } else {
    struct timeval curr_time;
    do_gettimeofday(&curr_time);
    printk("krpc: %ld.%06ld %.975s\n", curr_time.tv_sec, curr_time.tv_usec,
           message); /* note msgs are truncated at ~1000 chars to fit
                      * inside the 1024 printk limit imposed by the
                      * kernel */
  }

  va_end(ap);
}


/* rand-like interface to the kernel's random number generation function */
static int rand(void)
{
  int n;

  get_random_bytes(&n, sizeof(n));
  return n;
}


/************************ Radio Driving Functions **************************/

/* enqueue an incoming packet for app reading.  note that buffer is
 * considered to be the CALLER's buffer, we alloc our own new memory
 * for it. */
static void enqueue_incoming_packet(caddr_t buffer, int length,
                                    struct timeval packet_time)
{
  caddr_t new_buf;

  /* buffer full?  if so, just drop the packet. */
  if (RPC_INQ_FULL)
    return;

  /* we use the temp variable new_buf because kmalloc may block... i
   * don't *think* that this function needs to be reentrant, but
   * better safe than sorry */
  if ((new_buf = kmalloc(length, GFP_ATOMIC)) == NULL)
    return;

  /* put the new message on the tail of the incoming queue */
  memcpy(new_buf, buffer, length);
  rpc_inq[rpc_inq_tail].data = new_buf;
  rpc_inq[rpc_inq_tail].length = length;
  rpc_inq[rpc_inq_tail].time = packet_time;
  rpc_inq_tail = (rpc_inq_tail + 1) % RPC_BUFFER_SIZE;

  /* wake up processes that might be waiting to read */
  wake_up_interruptible(&rpc_read_wait);
}


/* enqueue a packet, assuming that the pointer we are given is memory
 * that has been malloc'd, and WE (not the caller) are responsible for
 * freeing it. */
static void enqueue_outgoing_packet(caddr_t message, int length)
{
  /* wait until the outgoing buffer isn't full */
  while (RPC_OUTQ_FULL) {
    interruptible_sleep_on(&rpc_write_wait);
  }
    
  rpc_outq[rpc_outq_tail].data = message;
  rpc_outq[rpc_outq_tail].length = length;
  rpc_outq_tail = (rpc_outq_tail + 1) % RPC_BUFFER_SIZE;

  /* now call the rpc timer handler function, so that if the radio is
   * idle, we can immediately transmit this packet */
  run_radio(RPC_WRITE);
}


/*
 * This macro is used to decide if the driver is "busy" (i.e., if the
 * periodic timer needs to be run or not).  We consider ourselves busy if
 * one of the following is true:
 * 1. someone has an RPC device open
 * 2. the write queue is not empty
 * 3. there is a write pending (should always be true if 2 is true) */
#define DRIVER_BUSY (open_files || !RPC_OUTQ_EMPTY || write_pending)


/* this function runs the radio at a high level: it checks for
 * incoming messages, if any, and moves outgoing messages out to the
 * radio.
 *
 * there are three reasons to run this function: 1) a write came from
 * userland that needs to be serviced; 2) an interrupt came from the
 * radio; 3) the RPC timer fired (every 10ms).  in all three cases,
 * this function is put on the immediate task queue to be run at a
 * "safe" time (search for where i call "run_radio" -- that function
 * schedules this function).
 *
 * note in particular that this function disables and re-enables
 * interrupts because it is not reentrant, so it is NOT safe to run
 * from within the interrupt handler itself.  this was the original
 * impetus for using the task queue instead of just making this
 * function the actual interrupt handler.
 *
 * aside from running from inside the interrupt handler, this function
 * was written to be "idempotent": i.e., safe for activation as
 * frequently as you want.  it can be triggered either by an interrupt
 * or by a timer.  interrupt-driven will give better performance
 * (lower latency for both sending and receiving), but it's nice to
 * also drive it via a timer whether or not you're expecting
 * interrupts to come in.  this way, the driver can take advantage of
 * but does not depend on interrupt-generating hardware.  also, it is
 * not affected by missed interrupts, except slight performance
 * degregation.  */
static void rpc_run_radio(void *data)
{
  static int message_count, tics_till_reset, tics_since_write_pending;
  static int radio_dead = 0;

  RDEBUG(50, "radio driver running for reason %d", (int) data);

  /* do nothing if there are no open files (or we're not initialized) */
  if (!DRIVER_BUSY)
    return;

  /* if we know the radio is dead go straight to the timeout section */
  if (radio_dead)
    goto timeout;

  /* if we are supposed to reset the radio, try resetting it */
  if (reset_needed) {
    reset_needed = 0;
    if (perform_reset_rpc() == RPC_TIMEOUT)
      goto timeout;
  }

  /* if a write has been pending for too long, time out */
  if (write_pending && tics_since_write_pending++ > RPC_TIMEOUT_TIME)
    goto timeout;

  /* disable interrupts: becase this function IS NOT REENTRANT and is
   * called by the interrupt handler!  danger will robinson! */
  pin_disable_irq();

  /* first try and receive a message.  this is always okay to do, even
   * if there is a write pending.  we don't care especially about the
   * return value (if a packet is waiting, rpc_incoming_upcall will be
   * up-called) */
  if (perform_receive_message() == RPC_TIMEOUT)
    goto timeout;

  /* forget the interrupt-arrival time -- if there wasn't a packet
   * waiting, the interrupt was signalling something else.  (if there
   * was a packet waiting, the timestamp has already been used in
   * rpc_incoming_upcall) */
  irq_time = unknown_time;

  /* now, if there's a write pending, try and send it IFF it will go
   * through immediately. */
  if (write_pending && perform_tx_poll()) {

    if (IN_HOLDOFF_PERIOD) {
      /* oops - a packet must have come in between the time we
       * requested the TX and the time the radio replied.  cancel the
       * transmission. */
      if (perform_tx_cancel() == RPC_TIMEOUT)
        goto timeout;
      write_pending = 0;
    } else {
      rpc_packet_t *p = &rpc_outq[rpc_outq_head];

      if (perform_send_message(p->data, p->length) == RPC_TIMEOUT)
        goto timeout;

      tx_frames++;
      tx_bytes += p->length;

      if (rpc_mode & RPC_MODE_COOKED) {
        RDEBUG(30, "SENDING RAW BYTES: (afa_id=%d) 0x%s",
               ((frag_common_header_t *)p->data)->frag_id,
               asc2hex(p->data, p->length));
      } else {
        RDEBUG(30, "SENDING RAW BYTES: 0x%s", asc2hex(p->data, p->length));
      }

      write_pending = 0;
      kfree(p->data);
      rpc_outq_head = (rpc_outq_head + 1) % RPC_BUFFER_SIZE;
      wake_up_interruptible(&rpc_write_wait);
    }
  }

  /* If we are a network interface: if the outgoing queue is empty,
   * and we are currently marked as busy, note that we are no longer
   * busy */
  if ((rpc_mode & RPC_MODE_NETIF) && RPC_OUTQ_EMPTY &&
      rpc_netif_dev != NULL && rpc_netif_dev->tbusy) {
    ((struct net_device_stats *)rpc_netif_dev->priv)->tx_packets++;
    rpc_netif_dev->tbusy = 0;
    mark_bh(NET_BH);
  }

  /* if there is something else to be transmitted, and we aren't
   * already in the middle of trying to transmit something, request a
   * TX from the RPC.  we'll check back with the rpc to see if it is
   * ready to receive our packet on the next timer tick or interrupt.
   * it's important to do this last (right before we re-enable
   * interrupts) so that we don't miss the interrupt. */
  if (!RPC_OUTQ_EMPTY && !write_pending && !IN_HOLDOFF_PERIOD) {
    perform_tx_request();
    write_pending = 1;
    tics_since_write_pending = 0;
  }

  pin_enable_irq();
  return;

 timeout:

  /* if the radio is dying for the first time, start the reset loop */
  if (!radio_dead) {
    radio_dead = 1;
    irq_time = unknown_time;
    write_pending = 0;
    message_count = 0;
    tics_till_reset = 0; /* try a reset immediately */
    RDEBUG(1, "radio appears dead, trying to reset...");
  }

  /* now try resetting the radio */
  if (tics_till_reset-- <= 0) {
    if (perform_reset_rpc() == RPC_TIMEOUT) {
      /* reset still failed... */
      tics_till_reset = RPC_RETRY_INTERVAL;
      if (++message_count == 5) {
        message_count = 0;
        RDEBUG(1, "radio dead, still trying...");
      }
    } else {
      /* reset was successful! */
      RDEBUG(1, "radio alive, reset successful");
      radio_dead = 0;
    }
  }
}


/* This is called when the timer fires; it just schedules rpc_run_radio */
void rpc_timer_handler(unsigned long data)
{
  run_radio(RPC_TIMER);

  /* decide if the driver still needs to be "on" -- i.e., if it is
   * still "busy" (see definition of "busy" above).  if so, reschedule
   * the timer.  if not, let the timer expire and turn off interrupts */
  if (!rpc_driver_on) {
    RDEBUG(0, "WHY DO I THINK THE RPC DRIVER IS OFF WHEN CLEARLY IT IS ON??");
  }

  if (DRIVER_BUSY) {
    /* driver should remain on */
    mod_timer(&rpc_timer, jiffies + RPC_POLL_INTERVAL);
  } else {
    /* driver should be turned off */
    RDEBUG(50, "Driver turning off");
    del_timer(&rpc_timer);
    pin_disable_irq();
    rpc_driver_on = 0;
  }
}


/* This is the interrupt handler... it just calls the same function as
 * used by the timer.  See the comment in the timer handler, above,
 * for a more detailed explanation. */
void rpc_interrupt_handler(int irq, void *dev, struct pt_regs *regs)
{
  do_gettimeofday(&irq_time);
  if (DRIVER_BUSY)
    run_radio(RPC_INTERRUPT);

  if (irq_debug)
    RDEBUG(0, "received an interrupt at %ld.%0ld", irq_time.tv_sec,
           irq_time.tv_usec);
}


/* This function is called by perform_receive_message() when a packet
 * is received by the RPC.  If the RPC is in raw mode, we just put it
 * right on the incoming (read) queue.  Otherwise, we call
 * process_incoming_fragment.  That function, in turn, will call the
 * deliver_packet_upcall function when reassembly is complete. */
void rpc_incoming_upcall(caddr_t buffer, int length)
{
  struct timeval packet_time;

  if (UNKNOWN_TIME(irq_time))
    do_gettimeofday(&packet_time); /* our best guess.. */
  else
    packet_time = irq_time;

  rx_frames++;
  rx_bytes += length;

  RDEBUG(30, "RECEIVED RAW BYTES: 0x%s", asc2hex(buffer, length));
  if (rpc_mode & RPC_MODE_COOKED) {
    /* in cooked mode with collision avoidance, decide when the next
     * time we are allowed to transmit is.  specifically, the current
     * time plus the holdoff time (randomized appropriately) */
    if (ca_active)
      rpc_holdoff_expiration = jiffies + ca_holdoff + (rand()%ca_holdoff_rand);

    process_incoming_fragment(buffer, length, packet_time);
  } else
    enqueue_incoming_packet(buffer, length, packet_time);
}


/* Fragment reassembly is complete.  If we are a file interface
 * device, enqueue this on the read queue.  If we are a network
 * interface, create an SKB and pass it up to the upper network
 * layers. */
void deliver_packet_upcall(caddr_t buffer, int length,
                           struct timeval packet_time)
{
  struct sk_buff *skb;

  if (rpc_mode & RPC_MODE_NETIF) {
    if ((skb = dev_alloc_skb(length+2)) == NULL) {
      RDEBUG(0, "rpc0 low on memory");
      return;
    }

    skb_reserve(skb, 2); /* align IP on 16b boundary */
    memcpy(skb_put(skb, length), buffer, length);
    skb->dev = rpc_netif_dev;
    skb->protocol = eth_type_trans(skb, rpc_netif_dev);
    skb->ip_summed = CHECKSUM_NONE;
    ((struct net_device_stats *)rpc_netif_dev->priv)->rx_packets++;
    netif_rx(skb);
  } else {
    enqueue_incoming_packet(buffer, length, packet_time);
  }
}


/* This is the upcall called by fragment_message() when it needs to
 * request a fragment be sent.  We malloc memory and put the newly
 * pointed-to data on the queue.  It is freed by the send routine as
 * it is dequeued. */
void send_fragment_upcall(caddr_t message, int length)
{
  caddr_t ptr = kmalloc(length, GFP_ATOMIC);
  memcpy(ptr, message, length);
  enqueue_outgoing_packet(ptr, length);
}


/* functions common to network interface open and file open: clear the
 * queue and start the polling timer */
void rpc_generic_open(void)
{
  perform_receive_message(); /* clear from the RPC's queue */
  rpc_inq_head = rpc_inq_tail = 0; /* clear our own queue */

  /* start the timer and interrupts firing, in case they already
   * aren't. */
  if (!rpc_driver_on) {
    RDEBUG(50, "Driver turning on");
    rpc_driver_on = 1; /* needs to be first due to race? */
    rpc_timer.expires = jiffies + RPC_POLL_INTERVAL;
    add_timer(&rpc_timer);
    pin_enable_irq();
  }
}


/******************* FILE INTERFACE FUNCTIONS ***************************/

static ssize_t rpc_read(struct file *file,
    char *user_buffer,    /* The buffer to fill with data */
    size_t user_length,   /* The length of the buffer */
    loff_t *offset)  /* Our offset in the file */
{
  int n, returned_packet_length, retval, peek;
  caddr_t returned_packet_data, tmp;
  struct timeval returned_packet_time;
  struct rpc_augmented *augmented;

  while (RPC_INQ_EMPTY) /* while incoming buffer is empty */ {
    if (file->f_flags & O_NONBLOCK)
      return -EAGAIN;

    /* sleep, waiting for a packet to arrive.  in most drivers this
     * would be a race condition (what if a packet arrives after the
     * comparison but before we go to sleep?).  in our case there is
     * not a race here (I THINK) because rpc_run_radio doesn't run
     * until the syscall exits -- it's on the immediate queue */
    interruptible_sleep_on(&rpc_read_wait);

    /* see if a signal woke us up */
    if (signal_pending(current))
      return -ERESTARTSYS;
  }

  /* Is the user requesting a peek?  (If so, remove that bit from the
   * length) */
  if (user_length & RPC_MSG_PEEK) {
    peek = 1;
    user_length &= ~RPC_MSG_PEEK;
  } else {
    peek = 0;
  }

  /* the following block, which pulls the first packet out of the head
   * of the read queue, must be atomic with respect to rpc_inq_head.
   * it's okay if the periodic handler runs here because that only
   * modifies rpc_inq_tail, not rpc_inq_head. */
  returned_packet_data = rpc_inq[rpc_inq_head].data;
  returned_packet_length = rpc_inq[rpc_inq_head].length;
  returned_packet_time = rpc_inq[rpc_inq_head].time;
  if (!peek)
    rpc_inq_head = (rpc_inq_head + 1) % RPC_BUFFER_SIZE;

  /* If the user has the device open in "augmented" mode, add the
   * extra data structure to the beginning of the packet. */
  if (rpc_mode & RPC_MODE_AUGMENTED) {
    /* first make room for the structure */
    tmp = kmalloc(returned_packet_length + sizeof(struct rpc_augmented),
                  GFP_KERNEL);
    memcpy(tmp + sizeof(struct rpc_augmented), returned_packet_data,
           returned_packet_length);
    kfree(returned_packet_data);
    returned_packet_data = tmp;
    returned_packet_length += sizeof(struct rpc_augmented);
    augmented = (struct rpc_augmented *) returned_packet_data;

    /* now populate the structure */
    augmented->packet_time = returned_packet_time;
  }

  /* calculate how many bytes the user will actually get */
  n = MIN(user_length, returned_packet_length);

  /* another thread might run here if we block accessing userspace (in
   * copy_to_user).  this is why i remove the message from the read
   * queue before copying it back to the user -- to preserve the
   * atomicity of those statements.  otherwise, a block here might
   * cause the same data to be returned to two different threads
   * trying to read from the same device.
   *
   * note that this function is reentrant by virtue of the fact that
   * each thread has its own stack, and we keep the message
   * temporarily buffered in local variables (i.e. on the stack) */
  if (copy_to_user(user_buffer, returned_packet_data, n) > 0)
    retval = -EFAULT;
  else
    retval = n;
  kfree(returned_packet_data);
  return retval;
}


static ssize_t rpc_write(struct file *file,
    const char *buffer,
    size_t length,
    loff_t *offset)
{
  caddr_t message = NULL;
  int max_length;

  if (length < 0)
    return -EINVAL;

  /* make sure the packet is not too long */
  if (rpc_mode & RPC_MODE_COOKED)
    max_length = MAX_PACKET_SIZE;
  else
    max_length = 27;
  if (length > max_length)
    return -EINVAL;

  /* make sure we have the memory for it */
  if ((message = kmalloc(length, GFP_KERNEL)) == NULL)
    return -ENOMEM;

  /* try to copy from user-space; abort if there's an error */
  if (copy_from_user(message, buffer, length) > 0) {
    kfree(message);
    return -EFAULT;
  }

  /* another thread might run here if we end up blocking on the
   * user-space access, so it's important for reentrancy that the
   * "message" variable is local.  see note about atomicity and
   * reentrancy in rpc_read. */
  if (rpc_mode & RPC_MODE_COOKED) {
    fragment_message(message, length);
    kfree(message);
  } else {
    enqueue_outgoing_packet(message, length);
    /* if radio is idle, write will start IMMEDIATELY... see
     * enqueue_outgoing_packet for details. */
  }

  return length;
}


static unsigned int rpc_poll(struct file *file, poll_table *wait)
{
  poll_wait(file, &rpc_read_wait, wait);
  if (!RPC_INQ_EMPTY) { /* incoming buffer not empty */
    return POLLIN | POLLRDNORM;
  }

  return 0;
}


static int rpc_open(struct inode *inode, struct file *file)
{
  int minor = MINOR(inode->i_rdev);

  if (!(open_files++)) {
    /* this is the first reader: clear the queue and set the periodic
     * timer for our housekeeping */
    rpc_mode = minor;
    rpc_generic_open(); /* open_files must be incremented before calling */
  } else {  /* this is one of the later files */
    /* we can only use one personality at a time! */
    if (minor != rpc_mode) {
      open_files--;
      return -EAGAIN;
    }
  }

  MOD_INC_USE_COUNT;
  return 0;
}

static int rpc_release(struct inode *inode, struct file *file)
{
  --open_files;

  if (!open_files) {
    /* this is the last to leave: housekeeping */

  }

  MOD_DEC_USE_COUNT;
  return 0;
}


/****************** Network Interface Functions **************************/


static int rpc_net_config(struct device *dev, struct ifmap *map)
{
  return 0;
}

static struct net_device_stats *get_stats(struct device *dev)
{
  return (struct net_device_stats *)dev->priv;
}

static int rpc_net_open(struct device *dev)
{
  /* don't use this as a netdevice if people are also using it as a file */
  if (open_files)
    return -EBUSY;

  if (dev == NULL || dev != rpc_netif_dev) {
    RDEBUG(0, "dev is null?  very strange... rpc driver is confused");
    return -ENODEV;
  }

  open_files++;
  dev->start = 1;
  dev->tbusy = 0;
  rpc_mode = RPC_MODE_COOKED | RPC_MODE_NETIF;
  rpc_generic_open();
  MOD_INC_USE_COUNT;
  return 0;
}

static int rpc_net_stop(struct device *dev)
{
  dev->start = 0;
  dev->tbusy = 1;
  open_files--;
  MOD_DEC_USE_COUNT;
  return 0;
}


static int rpc_net_tx(struct sk_buff *skb, struct device *dev)
{
  if (skb == NULL || dev == NULL || dev != rpc_netif_dev)
    return 0;

  dev->tbusy = 1;
  fragment_message(skb->data, skb->len);

  /* release the SKB -- fragment_message calls an upcall that
   * allocates its own memory for the RPC queue */
  dev_kfree_skb(skb);
  return 0;
}


static int rpc_net_ioctl(struct device *dev, struct ifreq *rq, int cmd)
{
  return 0;
}


int rpc_net_init(struct device *dev)
{
  int i;

  ether_setup(dev);

  dev->open = rpc_net_open;
  dev->stop = rpc_net_stop;
  dev->set_config = rpc_net_config;
  dev->hard_start_xmit = rpc_net_tx;
  dev->do_ioctl = rpc_net_ioctl;
  dev->get_stats = get_stats;
  dev->mtu = FRAG_DATA_LEN * 20;

  /*  dev->flags |= IFF_NOARP; */

  /* pick a random hardware address, but with the "local" bit set */
  dev->dev_addr[0] = 0x2; /* ethernet's "local address" bit */
  for (i = 1; i < ETH_ALEN; i++)
    dev->dev_addr[i] = rand() % 256;

  dev->priv = kmalloc(sizeof(struct net_device_stats), GFP_KERNEL);
  if (dev->priv == NULL)
    return -ENOMEM;
  memset(dev->priv, 0, sizeof(struct net_device_stats));

  return 0;
}


/***************** /proc Filesystem Handling Functions ******************/

#ifdef RPCPROC
/* function that handles writes to proc; i.e. setting a paramter */
static int rpc_write_proc(struct file *file, const char *buffer,
			  unsigned long count, void *data)
{
  char local[128];
  int num;
  int *param = (int *) data;

  if (count > 100)
    return -EINVAL;

  memcpy(local, buffer, count);
  local[count] = '\0';
  num = simple_strtoul(local, NULL, 10);
  *param = num;
  return count;
}

/* function that handles reads from proc; i.e. querying a parameter */
static int rpc_read_proc(char *page, char **start, off_t off,
                                         int count, int *eof, void *data)
{
  int *param = (int *) data;

  *start = 0;
  *eof = 1;
  return sprintf(page, "%d\n", *param);
}


/* this creates a proc entry that's an interface to our module parameters.
 * proc_base should be set first. 
 */
static void rpc_proc_register(char *name, int *parameter)
{
  struct proc_dir_entry *p;
  char *buf;

  if (num_rpc_proc_entries >= MAX_RPC_PROC_ENTRIES) {
    RDEBUG(0, "too many /proc/krpc entries!");
    return;
  }

  if ((buf=kmalloc(strlen(name) + strlen(KRPC_PROC_BASENAME) + 5, GFP_KERNEL))
      == NULL) {
    RDEBUG(0, "out of memory on /proc registration");
    return;
  }

  sprintf(buf, "%s/%s", KRPC_PROC_BASENAME, name);

  if ((p = create_proc_entry(buf, 0, NULL)) == NULL) {
    RDEBUG(1, "unable to create /proc/%s", buf);
    kfree(buf);
    return;
  }

  p->write_proc = rpc_write_proc;
  p->read_proc = rpc_read_proc;
  p->data = parameter;
  rpc_proc_entries[num_rpc_proc_entries++] = buf;
}


static void rpc_proc_init()
{
  if (create_proc_entry(KRPC_PROC_BASENAME, S_IFDIR,NULL) == NULL) {
    RDEBUG(1, "unable to create /proc entries");
    return;
  }

  rpc_proc_register("irq_debug", &irq_debug);
  rpc_proc_register("krpc_debug_level", &krpc_debug_level);
  rpc_proc_register("ca_active", &ca_active);
  rpc_proc_register("ca_holdoff", &ca_holdoff);
  rpc_proc_register("ca_holdoff_rand", &ca_holdoff_rand);
  rpc_proc_register("tx_frames", &tx_frames);
  rpc_proc_register("tx_bytes", &tx_bytes);
  rpc_proc_register("rx_frames", &rx_frames);
  rpc_proc_register("rx_bytes", &rx_bytes);
}

static void rpc_proc_cleanup()
{
  int i;

  for (i = 0; i < num_rpc_proc_entries; i++) {
    remove_proc_entry(rpc_proc_entries[i], NULL);
    kfree(rpc_proc_entries[i]);
  }
  num_rpc_proc_entries = 0;

  remove_proc_entry(KRPC_PROC_BASENAME, NULL);
}

#endif

/******************** Module-Related Functions ***************************/


static struct file_operations rpc_fops = {
  NULL, /* lseek */
  rpc_read, /* read */
  rpc_write, /* write */
  NULL, /* readdir */
  rpc_poll, /* poll */
  NULL, /* ioctl */
  NULL, /* mmap */
  rpc_open, /* open */
  NULL, /* flush */
  rpc_release, /* release */
  NULL, /* fsync */
  NULL  /* fasync */
};



int radiometrix_rpc_init(void)
{
  int retval;
  int error_return = -EIO;

  RDEBUG(1, "$Revision: 1.39 $, $Date: 2001/03/12 18:46:19 $, debuglevel=%d",
         krpc_debug_level);

#ifdef RPCPROC
  /* register our entries in /proc */
  rpc_proc_init();
#endif

  /* set up the timer (but don't schedule it yet) */
  init_timer(&rpc_timer);
  rpc_timer.function = rpc_timer_handler;
  rpc_timer.data = (unsigned long) 0;
  rpc_timer.expires = 0;

  /* set up the task queue used by the interrupt handler */
  rpc_queue.next = NULL;
  rpc_queue.sync = 0;
  rpc_queue.routine = rpc_run_radio;

  /* initialize the time structure that records when interrupts arrive */
  unknown_time.tv_sec = 0;
  unknown_time.tv_usec = 0;
  irq_time = unknown_time;

  /* get the par port using the parapin library */
  if ((retval = pin_init_kernel(lpt, rpc_interrupt_handler)) < 0) {
    error_return = retval;
    goto pin_error;
  }

  if (pin_have_irq())
    RDEBUG(1, "using interrupts from parallel port");
  else
    RDEBUG(1, "polling only (no interrupts)");

  /* Remember that we still need to reset the radio... */
  reset_needed = 1;

  /* initialize the fragmentation code */
  fragmentation_init();

  /* register our major number */
  if (register_chrdev(krpc_major_number, "rpc", &rpc_fops) < 0) {
    RDEBUG(0, "unable to register character device");
    goto chrdev_error;
  }
  RDEBUG(1, "registered char device major number %d", krpc_major_number);

  /* register the network device */
  memcpy(rpc_netif_dev->name, "rpc0", 5);
  if ((retval = register_netdev(rpc_netif_dev)) < 0) {
    RDEBUG(0, "error %d registering device %s", retval, rpc_netif_dev->name);
    goto net_error;
  }
  RDEBUG(1, "registered network device %s", rpc_netif_dev->name);

  return 0;


 net_error:
  unregister_chrdev(krpc_major_number, "rpc");
 chrdev_error:
  fragmentation_shutdown();
  pin_release();
 pin_error:
  return error_return;
}


#ifdef MODULE

int init_module(void)
{
  return radiometrix_rpc_init();
}

void cleanup_module(void)
{
  RDEBUG(1, "shutting down");

  /* stop the timer */
  if (rpc_driver_on) {
    del_timer(&rpc_timer);
    pin_disable_irq();
  }

  /* free memory used by the frag code */
  fragmentation_shutdown();

  /* unregister the character device */
  unregister_chrdev(krpc_major_number, "rpc");

  /* unregister the network device */
  unregister_netdev(rpc_netif_dev);
  if (rpc_netif_dev->priv != NULL)
    kfree(rpc_netif_dev->priv);

#ifdef RPCPROC
  /* unregister the /proc entries */
  rpc_proc_cleanup();
#endif

  /* release the parallel port */
  pin_release();
}
#endif
