#define DEBUG(x, y)
#define DEBUG3(x, y, z)

#define __LIBRPC_C__

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>
#include <sys/time.h>
#include <setjmp.h>

#include "rpc.h"
#include "rpc_lowlevel.h"
#include "parapin.h"

int sig_caught = 0;
int busy = 0;
sigjmp_buf env;

incoming_message_callback_prototype incoming_message_callback = NULL;
termination_callback_prototype termination_callback = NULL;

/***********************************************************************/

/* initialize the RPC... assume for now that port = LPT1.  Returns 0
 * on success, -1 on error. */
int rpc_init()
{
  /* Initialize the underlying "parapin" library */
  if (pin_init_user(LPT1) < 0)
    return -1;

  /* Set up signal handlers */
  signal(SIGINT, sighandler);
  signal(SIGTERM, sighandler);
  signal(SIGHUP, sighandler);

  /* Reset the RPC */
  rpc_reset();

  return 0;
}


/* Try and reset the RPC until it succeeds. */
void rpc_reset()
{
  int reset_retried = 0;

  /* if the user is trying to bail, don't bother trying to reset */
  if (sig_caught)
    terminate();

  while (perform_reset_rpc() == RPC_TIMEOUT) {
    reset_retried++;
    fprintf(stderr, "timeout %d while trying to reset RPC... still trying\n",
            reset_retried);
    sleep(4);
  }

  if (reset_retried)
    fprintf(stderr, "RPC reset successfully\n");
}


void rpc_register_incoming_message_callback(incoming_message_callback_prototype x)
{
  incoming_message_callback = x;
}


void rpc_register_termination_callback(termination_callback_prototype x)
{
  termination_callback = x;
}


/* This is a wrapper around the real receive_message in case we
 * receive a control-c in the middle of processing, and to implement a
 * timeout in case the radio dies.  Return value of 1 means a message
 * was waiting; 0 means no message was waiting.  Return value of 1
 * (message received) also means we called the user-supplied callback
 * if any. */
int rpc_receive_message()
{
  int retval;

  busy = 1;
  retval = perform_receive_message();
  busy = 0;

  if (retval == RPC_TIMEOUT) {
    retval = 0;
    rpc_reset();
  }

  if (sig_caught)
    terminate();

  return retval;
}



/* This is a wrapper around the real send_message in case we receive a
 * control-c in the middle of processing. */
int rpc_send_message(caddr_t message, int length)
{
  int retval;

  busy = 1;
  retval = perform_send_message(message, length);
  busy = 0;

  if (retval == RPC_TIMEOUT) {
    retval = 0;
    rpc_reset();
  }

  if (sig_caught)
    terminate();

  return retval;
}


void rpc_send_control(char address, char value)
{
  int retval ;

  busy = 1;
  retval = perform_send_control(address, value);
  busy = 0;

  if (retval == RPC_TIMEOUT)
    rpc_reset();

  if (sig_caught)
    terminate();
}



/******************** Private Functions ********************************/

/* signal handler.  if we are currently in the middle of a critical
 * section, remember that we should terminate.  if not, terminate
 * immediately. */
static void sighandler()
{
  if (busy)
    sig_caught = 1;
  else
    terminate();
}


/* Terminate the program - first calling the user-supplied termination
 * function if necessary. */
static void terminate()
{
  if (busy) {
    fprintf(stderr, "HUH??  terminate called at the wrong time!\n");
    return;
  }

  if (termination_callback != NULL)
    termination_callback();

  exit(0);
}


/* this upcall is invoked by rpc_lowlevel when a message is received
 * inside of perform_receive_message() */
void rpc_incoming_upcall(caddr_t message, int length)
{
  if (incoming_message_callback != NULL)
    incoming_message_callback(message, length);
}
