/*
 * Apitest: Sample use of radiometrix api.  Simply sends a packet to
 * the server once every second, and, every second, prints any packets
 * received.  Takes two arguments: first, the name of the debugging
 * server to use; second, the contents of the message.
 */

#define MAX_MSG_SIZE   512

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <string.h>
#include <assert.h>
#include <signal.h>

#include "rpcd_api.h"


int incoming_sock;
char *outgoing_message;

static void listen_on_port(int port)
{
  struct linger ld;
  struct sockaddr_in sa;

  /* Create the incoming socket */
  if ((incoming_sock = socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP)) < 0) {
    perror("Error creating socket");
    exit(1);
  }

  /* Turn off lingering in case the process hangs */
  ld.l_onoff = 0;
  ld.l_linger = 0;
  if (setsockopt(incoming_sock, SOL_SOCKET, SO_LINGER, (caddr_t)&ld,
                 sizeof(ld))<0)
    perror("setsockopt SO_LINGER");

  /* Clear the structure */
  memset((caddr_t)&sa, 0, sizeof(sa));

  /* internet address on specified port, accept on any interface */
  sa.sin_family = AF_INET;
  sa.sin_port = htons(port);
  sa.sin_addr.s_addr = htonl(INADDR_ANY);

  fprintf(stderr, "app: trying to bind to port %d...\n", port);

  if (bind(incoming_sock, (struct sockaddr *) &sa, sizeof(sa)) < 0) {
    perror("bind failed");
    close(incoming_sock);
    exit(1);
  }

  fprintf(stderr, "app: bind succeeded on port %d\n", port);
}



/* read one packet off the wire and print it out */
void read_packet()
{
  char buf[MAX_MSG_SIZE];
  int bytes_read;

  bytes_read = recvfrom(incoming_sock, buf, MAX_MSG_SIZE, 0, NULL, 0);

  if (bytes_read < 0) {
    perror("app: error reading from network");
    /*    exit(1); */
  } else if (bytes_read == 0) {
    fprintf(stderr, "app: whoa!  recvfrom returned 0??\n");
    /*     exit(1); */
  } else {
    buf[bytes_read] = '\0';
    printf("**************** APP received: %s\n", buf);
  }
}


/* Send a message using the radiometrix library, and re-set the alarm
 * so we send another packet one second from now. */
void send_packet(int signal)
{
  alarm(2);
  printf("**************** APP sending: %s\n", outgoing_message);
  radiometrix_send(outgoing_message, strlen(outgoing_message));
}


int main(int argc, char *argv[])
{

  /* Make sure we have the arguments we need */
  if (argc != 1 && argc != 2) {
    fprintf(stderr, "app: Usage: %s [test-message]\n", argv[0]);
    exit(1);
  }

  radiometrix_set_debug_level(10);

  /* Initialize the radiometrix library */
  if (radiometrix_init() < 0) {
    fprintf(stderr, "app: radiometrix library failed - exiting\n");
    exit(1);
  }

  /* Start listening on the diffusion port */
  listen_on_port(DIFFUSION_PORT);

  /* Optionally send a message every 2 seconds. */
  if (argc == 2) {
    outgoing_message = argv[1];
    signal(SIGALRM, send_packet);
    alarm(2);
  }

  /* Now, loop forever: check for incoming packets, time out, send our
   * own packet.  First set up the signal handler. */
  do {
    read_packet();
  } while (1);


  fprintf(stderr, "app: exiting[%d]\n", getpid());
}
