/*
 * RPC daemon: a radiometrix device driver, used so that multiple
 * processes can send and receive messages on a single radiometrix
 * device.
 */

#define __DAEMON_C__
#define MAX_FDS 256

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <sys/wait.h>
#include <sys/time.h>

#include "rpcd_api.h"
#include "rpcd.h"
#include "rpc.h"
#include "afa_fragmentation.h"

int server_fd, incoming_frag_fd, delivery_fd;
struct sockaddr_in delivery_sa;

int num_clients;
client_t client[MAX_FDS];

int radiometrix_mode = NORMAL;
char *debug_hostname = NULL;
int be_quiet = 0;

struct timeval dummy_time;

/****************************************************************************/


void run_daemon()
{
  fragmentation_init();

  /* initialize structures etc. */
  num_clients = 0;
  srand(time(0) ^ ((getpid() & 0xff) << 24));

  /* listen for connections from clients */
  if ((server_fd = server_setup(RPCLIB_INTERNAL_PORT, TCP)) < 0) {
    RAW_RDEBUG(0, perror("RPC daemon unable to open server socket"));
    exit(1);
  }

  /* make ONLY the server fd nonblocking (so that accept() does not block) */
  if (nonblock(server_fd) < 0) {
    RAW_RDEBUG(1, perror("setting server_fd nonblock"));
    exit(1);
  }

  /* set up socket we can use to deliver received packets to localhost */
  if ((delivery_fd =
       create_udp_source("localhost", DIFFUSION_PORT, &delivery_sa)) < 0) {
    RDEBUG(1, "unable to create port for making local deliveries!");
    exit(1);
  }

  /* get ready to receive incoming fragments */
  if (radiometrix_mode == FRAG_DEBUG) {

    /* first set up conn for outgoing fragments */
    if (init_server_connection(debug_hostname, DIFFUSION_PORT) < 0) {
      RAW_RDEBUG(0, perror("RPC daemon unable to set up outgoing frag port"));
      exit(0);
    }
    if ((incoming_frag_fd = server_setup(RPC_FRAG_DEBUG_PORT, UDP)) < 0) {
      RAW_RDEBUG(0, perror("RPC daemon unable to listen to frag port"));
      exit(0);
    }
    nonblock(incoming_frag_fd);
  } else {
    /* initialize the low-level radiometrix library */
    if (rpc_init() < 0) {
      RDEBUG(0, "unable to initialize low-level RPC library - exiting");
      exit(1);
    }
    rpc_register_incoming_message_callback(process_incoming_fragment_wrapper);
  }

  /* Now infinitely loop - listening for messages from children and RPC */
  for (;;) {
    fd_set read_set;
    int i;

    /* see if new clients are trying to connect */
    accept_new_clients();

    /* see if anyone has sent us a message that needs to go to the radio */
    if (check_client_messages(&read_set) > 0) {
      for (i = 0; i < num_clients; i++)
        if (FD_ISSET(client[i].fd, &read_set))
          process_outgoing_message(i);
    }

    /* purge clients that might be dead from previous read */
    purge_client_array();

    /* check for incoming fragments coming from the radio or
     * frag-debug server */
    check_incoming_fragments();
  }
}


/* added 5 july 2000 because the fragmentation interface changed to
 * include packet times but i don't actually want to implement that
 * in userspace because no one uses it
 */
void process_incoming_fragment_wrapper(caddr_t message, int length)
{
  process_incoming_fragment(message, length, dummy_time);
}

/*
 * If there are any new connections that clients are trying to
 * establish with us, accept the stream and record the fd in our
 * client_fd array.
 */
static void accept_new_clients()
{
  int fd;

  if ((fd = accept(server_fd, NULL, NULL)) >= 0) {
    client[num_clients].fd = fd;
    client[num_clients].dead = 0;
    num_clients++;
    return;
  } else if (fd < 0 && errno != EAGAIN) {
    RAW_RDEBUG(1, perror("accept"));
    exit(1);
  }
}


/* Remove clients marked as dead from the client array. */
static void purge_client_array()
{
  int in, out = 0;

  /* Now delete the dead clients from the client[] array */
  for (in = 0, out = 0; in < num_clients; in++) {
    if (!client[in].dead) {
      /* Client is not dead -- keep it in the array */
      if (in != out)
        client[out] = client[in];
      out++;
    } else {
      /* Client is dead - close their descriptor */
      close(client[in].fd); 
      RDEBUG(2, "Disconnecting client");
    }
  }
  num_clients = out;
}


/*
 * Use select to get a list (returned in "read_set") of all
 * descriptors that are readable.  This may include both real messages
 * and descriptors that have closed.
 */
static int check_client_messages(fd_set *read_set)
{
  int max = 0, retval, i;
  struct timeval t;

  /* set up the select timeout */
  t.tv_sec = 0;
  t.tv_usec = SELECT_TIME;

  /* set up the read set */
  FD_ZERO(read_set);
  for (i = 0; i < num_clients; i++) {
    FD_SET(client[i].fd, read_set);
    if (client[i].fd > max)
      max = client[i].fd;
  }

  /* determine which FDs are readable */
  if ((retval = select(max+1, read_set, NULL, NULL, &t)) < 0) {
    RAW_RDEBUG(1, perror("select"));
    exit(1);
  } else {
    return retval;
  }
}


/* check for incoming fragments coming from the radio or frag-debug
 * server */
void check_incoming_fragments()
{
  char buf[128];
  int len, iterations, idle;

  switch (radiometrix_mode) {
  case NORMAL:
  case DEBUG:
    /* keep prodding the RPC library to deliver messages until there
     * are none left */
    if (rpc_receive_message()) {
      RDEBUG(8, "going into active receive mode");
      iterations = idle = 0;
      while (/* iterations <= 150000 &&  */ idle <= 15000) {
        iterations++;
        idle++;
        if (rpc_receive_message())
          idle = 0;
      }
      RDEBUG(8, "leaving active receive mode");
    }
    break;
  case FRAG_DEBUG:
    /* try and receive a fragment off the network */
    if ((len = recvfrom(incoming_frag_fd, buf, 128, 0, NULL, 0)) > 0)
      process_incoming_fragment(buf, len, dummy_time);
    break;
  default:
    RDEBUG(1, "huh?  unknown mode in daemon message receive loop");
    break;
  }
}



/*********************** fragmentation routines ****************************/


/*
 * process an incoming message.  Note that the argument is NOT an fd,
 * but rather an index into the client_fd[] array.
 */
static void process_outgoing_message(int fd_index)
{
  internal_comm_hdr hdr;
  caddr_t message;

  hdr.length = 0;

  /* make sure the FD is valid */
  if (fd_index < 0 || fd_index >= num_clients) {
    RDEBUG(1, "process_outgoing_message: WHOA... screwed up index.");
    return;
  }

  /* read the "internal comm" header off the stream */
  if (stream_read(client[fd_index].fd, (caddr_t) &hdr, sizeof(hdr)) < 0) {
    client[fd_index].dead = 1;
    return;
  }

  /* allocate memory for the incoming message */
  if ((message = malloc(hdr.length)) == NULL) {
    RDEBUG(0, "yikes!  out of memory in process_outgoing_message - giving up");
    exit(1);
  }

  /* read the actual (post-header) message into mem we just allocated */
  if (stream_read(client[fd_index].fd, message, hdr.length) < 0) {
    client[fd_index].dead = 1;
    return;
  }

  /* debugging: print the message */
  if (radiometrix_debug_level >= 3) {
    char buf[128];

    sprintf(buf, "GOT MESSAGE (length %d): ", hdr.length);
    rdebug_print_message(3, buf, message, hdr.length);
  }

  /* fragment the message into bite-size chunks and send to radio */
  fragment_message(message, hdr.length);

  /* free memory allocated above */
  free(message);
}



/* When we call fragment_message(), it will generate a series of
 * upcalls to this function (send_fragment_upcall) to deliver each
 * individual fragment. */
void send_fragment_upcall(caddr_t message, int length)
{
  if (radiometrix_mode == FRAG_DEBUG)
    send_to_server(message, length);
  else
    rpc_send_message(message, length);
  usleep(SENDER_IPG);
}


/* packet successfully grabbed from below.  pass up to the layer above. */
void deliver_packet_upcall(caddr_t message, int length,
                           struct timeval packet_time)
{
  rdebug_print_message(3, "REASSEMBLED: ", message, length);

  if (sendto(delivery_fd, message, length, 0,
             (struct sockaddr *) &delivery_sa, sizeof(delivery_sa)) < 0) {
    RAW_RDEBUG(1, perror("making a local delivery"));
  }
  
}



/**************************** Main *****************************************/


void print_usage(char *progname)
{
  fprintf(stderr, "usage: %s [-hq] [-d debug_level] [-s frag_debug_host]\n\n",
          progname);
  fprintf(stderr, "        -d: Debug level; default is %d\n",
          DEFAULT_RDEBUG_LEVEL);
  fprintf(stderr, "        -f: run in Foreground; do not detach from console\n");
  fprintf(stderr, "        -h: print this Help message\n");
  fprintf(stderr, "        -q: be very Quiet\n");
  fprintf(stderr, "        -s: enter fragServer debug mode, sending frags to frag_debug_host\n");

}



int main(int argc, char *argv[])
{
  extern int optind;
  extern int opterr;
  extern int optopt;
  extern char *optarg;
  int arg, tmp, need_usage = 0;

  rinit_debug(argv[0]);

  opterr = 0;

  while ((arg = getopt(argc, argv, "d:fhqs:")) != EOF) {
    switch (arg) {
    case 'd':

      if ((tmp = atoi(optarg)) < 0) {
        RDEBUG(1, "warning: -d flag with illegal debug level '%s'", optarg);
      } else {
        radiometrix_debug_level = tmp;
      }
      break;
    case 'f':
      radiometrix_mode = DEBUG;
      break;
    case 'h':
      need_usage = 1;
      break;
    case 'q':
      be_quiet = 1;
      break;
    case 's':
      radiometrix_mode = FRAG_DEBUG;
      radiometrix_debug_level = 10;
      debug_hostname = strdup(optarg);
      break;
    default:
      RDEBUG(1, "error: unrecognized switch '%c'", optopt);
      need_usage = 1;
      break;
    }
  }

  /* print help and exit if there was an error in the arguments */
  if (need_usage) {
    print_usage(argv[0]);
    exit(1);
  }

  /* If we're not running in the background, fork and disconnect from parent */
  if (radiometrix_mode == NORMAL) {
    int child_pid;

    child_pid = fork();
    if (child_pid > 0) {
      if (!be_quiet)
        fprintf(stderr, "radiometrix daemon started, pid %d\n", child_pid);
      exit(0);
    }
  } else {
    if (!be_quiet)
      fprintf(stderr, "radiometrix daemon started on console\n");
  }

  run_daemon();

  return 0;
}

