
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <fcntl.h>
#include <signal.h>
#include <ctype.h>
#include <limits.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <sys/uio.h>
#include <netinet/in.h>
#include <netdb.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "rpcd_api.h"
#include "rpcd.h"

#ifndef INADDR_NONE
#define INADDR_NONE  ((u_long) 0xffffffff)
#endif



/* Turns POSIX non-blocking on for a socket; returns 0 if OK, -1 if not */
int nonblock(int fd)
{
  int val;

  if ((val = fcntl(fd, F_GETFL, 0)) < 0)
    return val;
  val |= O_NONBLOCK;
  if (fcntl(fd, F_SETFL, val) < 0)
    return val;

  return 0;
}


/*
 * The following is the tcp_open() routine copied from Stevens' book,
 * UNIX Network Programming.  The source for the entire book is available
 * in an archive at ftp.uu.net.  I made some modifications to bring the
 * code more up-to-date (e.g. ANSI C function prototype, changing bzero()
 * to memset(), etc.) - J.E.
 */

/* The following globals are available to the caller, if desired. */
struct sockaddr_in tcp_srv_addr; /* server's Internet socket addr */
struct servent tcp_serv_info;    /* from getservbyname() */
struct hostent tcp_host_info;    /* from gethostbyname() */

/*
 * Arguments:
 *  host:    name or dotted-decimal addr of other system
 *  service: name of service being requested; can be NULL iff port > 0
 *  port:    if == 0, nothing special - use port# of service
 *           if < 0, bind a local reserved port
 *           if > 0, it's the port# of server (host-byte-order)
 *
 * Returns socket descriptor if OK, else -1 on error
 */
int tcp_open(char *host, char *service, int port)
{
  int fd = -1, resvport;
  unsigned long inaddr;
  struct servent *sp;
  struct hostent *hp;

  /* Initialize the server's Internet address structure.  We'll store the
   * actual 4-byte Internet address and the 2-byte port# below. */
  memset(&tcp_srv_addr, '\0', sizeof(tcp_srv_addr));
  tcp_srv_addr.sin_family = AF_INET;

  if (service != NULL) {
    if ((sp = getservbyname(service, "tcp")) == NULL) {
      RDEBUG(1, "tcp_open: unknown service: %s/tcp", service);
      return -1;
    }
    tcp_serv_info = *sp;        /* structure copy */
    if (port > 0)
      tcp_srv_addr.sin_port = htons(port); /* caller's value */
    else
      tcp_srv_addr.sin_port = sp->s_port;  /* service's value */
  } else {
    if (port <= 0) {
      RDEBUG(1, "tcp_open: must specify either service or port");
      return (-1);
    }
    tcp_srv_addr.sin_port = htons(port);
  }

  /* First try to convert the host name as a dotted-decimal number.
   * Only if that fails do we call gethostbyname(). */

  if ((inaddr = inet_addr(host)) != INADDR_NONE) {  /* it's dotted-decimal */
    memcpy(&tcp_srv_addr.sin_addr, &inaddr, sizeof(inaddr));
    tcp_host_info.h_name = NULL;
  } else {
    if ((hp = gethostbyname(host)) == NULL) {
      RDEBUG(1, "tcp_open: '%s' host name resolution error", host);
      return -1;
    }
    tcp_host_info = *hp;        /* found it by name, structure copy */
    memcpy(&tcp_srv_addr.sin_addr, hp->h_addr, hp->h_length);
  }

  if (port >= 0) {
    if ((fd = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
      RDEBUG(1, "tcp_open: can't create TCP socket");
      return -1;
    }
  } else if (port < 0) {
    resvport = IPPORT_RESERVED - 1;
    if ((fd = rresvport(&resvport)) < 0) {
      RDEBUG(1, "tcp_open: can't get a reserved TCP port");
      return -1;
    }
  }

  /* Connect to the server. */
  if (connect(fd, (struct sockaddr *) &tcp_srv_addr, sizeof(tcp_srv_addr))<0) {
    RDEBUG(4, "tcp_open: can't connect to server %.70s.%d", host, port);
    RAW_RDEBUG(4, perror("...error was:"));
    close(fd);
    return -1;
  }

  return fd;    /* all OK */
}



#define MAX_RETRIES 20

int tcp_open_retry(char *site, int port)
{
  int retries = 0, fd = -1, curr_sleep = 1;

  while (fd < 0 && retries <= MAX_RETRIES) {
    if ((fd = tcp_open(site, NULL, port)) < 0)
      RDEBUG(1, "connection to daemon failed, still trying...");
    sleep(curr_sleep);
    if ((curr_sleep < 32) && (retries++ & 1))
      curr_sleep *= 2;
  }

  return fd;
}



/* Set up a TCP server: create the socket, listen, etc. */
int server_setup(int server_port, int port_type)
{
  struct sockaddr_in serv_addr;
  int opt;
  int server_sock = -1;

  /* create the socket */

  if (port_type == TCP)
    server_sock = socket(PF_INET, SOCK_STREAM, 0);
  else if (port_type == UDP)
    server_sock = socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP);

  if (server_sock < 0) {
    RAW_RDEBUG(1, perror("Can't create socket"));
    return -1;
  }

  /*
   * if the OS supports reuseaddr, use it to prevent us from rebinding
   * to the same port later if a process is hanging around.
   */
#if defined(SO_REUSEADDR)
  opt = 1;
  if (setsockopt(server_sock, SOL_SOCKET, SO_REUSEADDR, (char *) &opt,
                 sizeof(opt)) < 0) {
    RAW_RDEBUG(1, perror("setsockopt REUSEADDR"));
  }
#endif

  /*
   * if the OS supports no-linger, use it to prevent the process from
   * hanging around and preventing us from rebinding to the port
   */
#if defined(SO_LINGER)
  {
    struct linger ld;
 
    ld.l_onoff = 0;
    ld.l_linger = 0;
    if (setsockopt(server_sock, SOL_SOCKET, SO_LINGER, (char *) &ld,
                   sizeof(ld)) < 0) {
      RAW_RDEBUG(1, perror("setsockopt LINGER"));
    }
  }
#endif

  /* set up the address structure - we are willing to accept on any iface */
  memset(&serv_addr, '\0', sizeof(serv_addr));
  serv_addr.sin_family = AF_INET;
  serv_addr.sin_addr.s_addr = htonl(INADDR_ANY);
  serv_addr.sin_port = htons(server_port);

  /* bind socket to address */
  if (bind(server_sock, (struct sockaddr *) &serv_addr, sizeof(serv_addr))<0){
    RAW_RDEBUG(1, perror("bind failed"));
    close(server_sock);
    return -1;
  }

  /* tell the OS we want to accept connections */
  if (port_type == TCP) {
    if (listen(server_sock, 5) < 0) {
      RAW_RDEBUG(1, perror("listen failed"));
      close(server_sock);
      return -1;
    }
  }

  return server_sock;
}


int create_udp_source(char *hostname, u_int16_t dest_port,
                      struct sockaddr_in *sa)
{
  struct in_addr ad;
  struct hostent *he;
  int fd;

  if (hostname == NULL || *hostname == '\0' || sa == NULL) {
    RDEBUG(1, "hostname and socket must be provided");
    return -1;
  }

  if ((fd = socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP)) < 0) {
    RAW_RDEBUG(1, perror("create socket"));
    return -1;
  }

  memset(sa, 0, sizeof(struct sockaddr_in));
  sa->sin_family = AF_INET;
  sa->sin_port = htons(dest_port);

  
  if (inet_aton(hostname, &ad) != 0) {
    /* valid IP address given */
   sa->sin_addr = ad;
  } else {
    /* no valid IP address given; try resolving hostname */
    if ((he = gethostbyname(hostname)) == NULL) {
      RDEBUG(1, "invalid hostname %s", hostname);
      exit(1);
    }
    if (he->h_addr_list == NULL || he->h_addr_list[0] == NULL) {
      RDEBUG(1, "strange hostname %s", hostname);
      exit(1);
    }
    memcpy(&(sa->sin_addr.s_addr), he->h_addr_list[0], he->h_length);
  }

  return fd;
}


/* 
 * stream_read: reads "nbytes" octets from stream "fd" and writes it
 * to the buffer at "ptr".  Repeatedly calls write in case the entire
 * buffer is not accepted due to a blocked stream.
 *
 * NOTE: This version of stream_write assumes that a fd has *NOT* been
 * set to be non-blocking.  It will block until the entire read is
 * complete by executing multiple reads if the reads become blocked.
 *
 * Returns:
 * 0    if the read was successful, all bytes read as requested
 * -1   if there was some error forcing a close, including EOF on stream
 */
int stream_read(int fd, caddr_t ptr, size_t nbytes)
{
  size_t to_read = nbytes;
  int n;

  /* Read until we have all we need... */
  while (to_read) {
    n = read(fd, ptr, to_read);
    if (n == 0)
      return -1; /* connection reset by peer */
    else if (n < 0) {
      /* retval of the read indicates an error */
      RDEBUG(1, "stream read() on fd %d failed...", fd);
      RAW_RDEBUG(1, perror("...error was"));
      return -1;
    }

    to_read -= n;
    ptr += n;
  }

  return 0;
}



/* 
 * stream_write: writes "nbytes" octets to stream "fd" from buffer at
 * "ptr".  Repeatedly calls write in case the entire buffer is not
 * accepted due to a blocked stream.
 *
 * NOTE: THIS VERSION OF stream_write ASSUMES THAT A FD HAS *NOT* BEEN
 * SET TO BE NON-BLOCKING.
 *
 * Returns:
 * 0    if the write was succesful
 * -1 if there was some error forcing a close
 */
int stream_write(int fd, caddr_t ptr, size_t nbytes)
{
  int n;

  if (ptr == NULL)
    return -1;

  while (nbytes) {
    n = write(fd, ptr, nbytes);
    if (n < 0) {
      RDEBUG(1, "stream write() on fd %d failed", fd);
      RAW_RDEBUG(1, perror("...error was"));
      return -1;
    }
    nbytes -= n;
    ptr += n;
  }

  return 0;
}


/*
 * Write one or more buffers to descriptor "fd"; buffers and lengths
 * defined the same way as writev(2).
 *
 * NOTE: THIS VERSION OF stream_write ASSUMES THAT A FD HAS *NOT* BEEN
 * SET TO BE NON-BLOCKING.
 *
 * Returns:
 * 0    if the write was succesful
 * -1 if there was some error forcing a close
 */
int multi_stream_write(int fd, struct iovec *iov, int iovcnt)
{
  int n;

  for (;;) {
    if (iovcnt <= 0) {          /* Negative buffers to write?? (!) */
      RDEBUG(1, "warning: iovcnt <= 0 in multi_stream_write()!");
      return -1;
    } else if (iovcnt == 0)     /* No buffers left to write */
      return 0;
    else if (iovcnt == 1)       /* One buffer left to write */
      return stream_write(fd, iov[0].iov_base, iov[0].iov_len);
    else {                      /* More than one buffer to write */

      /* Write as many buffers as possible */
      n = writev(fd, iov, iovcnt);

      /* Check and see if a permanent error was reported */
      if (n < 0) {
        RDEBUG(1, "multi-stream writev() on fd %d failed", fd);
        RAW_RDEBUG(1, perror("...error was"));
        return -1;
      }

      /* Calculate the new iov and iovcnt for the next write (n is the number
       * of bytes written by the previous write) */
      while (n && iovcnt > 0) {
        if (iov[0].iov_len == n && iovcnt == 1) {
          /* We wrote the last buffer successfully: we're done */
          return 0;
        } else if (iov[0].iov_len <= n) {
          /* We completely wrote iov[0]: skip to the next iov */
          n -= iov[0].iov_len;
          iov++;
          iovcnt--;
        } else {
          /* We stopped writing in the middle of this iov */
          iov[0].iov_base += n;
          iov[0].iov_len -= n;
          n = 0;
        }
      }
    }
  }
}



/**********************************************************************/

/* FUNCTION BELOW WAS TAKEN FROM TCPDUMP v3.4 */

/*
 * Copyright (c) 1997
 *      The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that: (1) source code distributions
 * retain the above copyright notice and this paragraph in its entirety, (2)
 * distributions including binary code include the above copyright notice and
 * this paragraph in its entirety in the documentation or other materials
 * provided with the distribution, and (3) all advertising materials mentioning
 * features or use of this software display the following acknowledgement:
 * ``This product includes software developed by the University of California,
 * Lawrence Berkeley Laboratory and its contributors.'' Neither the name of
 * the University nor the names of its contributors may be used to endorse
 * or promote products derived from this software without specific prior
 * written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

/*
 * Returns the difference between gmt and local time in seconds.
 * Use gmtime() and localtime() to keep things simple.
 */
int32_t
gmt2local(time_t t)
{
        register int dt, dir;
        register struct tm *gmt, *loc;
        struct tm sgmt;

        if (t == 0)
                t = time(NULL);
        gmt = &sgmt;
        *gmt = *gmtime(&t);
        loc = localtime(&t);
        dt = (loc->tm_hour - gmt->tm_hour) * 60 * 60 +
            (loc->tm_min - gmt->tm_min) * 60;

        /*
         * If the year or julian day is different, we span 00:00 GMT
         * and must add or subtract a day. Check the year first to
         * avoid problems when the julian day wraps.
         */
        dir = loc->tm_year - gmt->tm_year;
        if (dir == 0)
                dir = loc->tm_yday - gmt->tm_yday;
        dt += dir * 24 * 60 * 60;

        return (dt);
}


void *check_malloc(size_t size)
{
  void *ptr;

  if ((ptr = malloc(size)) == NULL) {
    fprintf(stderr, "Malloc failed - out of memory?");
    exit(1);
  }
  return ptr;
}

