/*
 * Various network-related utilities used by different modules.
 *
 * $Id: util.c,v 1.2 2000/06/27 23:17:13 jelson Exp $
 *
 * $Log: util.c,v $
 * Revision 1.2  2000/06/27 23:17:13  jelson
 * Revamped the debugging infratstructure -- I think it may need more work,
 * but it's better than the mess that it was.  I am trying to standardize on
 * the RDEBUG(message_level, printf_format, args...) interface; the macro
 * prevents anything from being evaluated if the message level is not high
 * enough.  I prefer this to a function call because it prevents arguments
 * from being evaluated.
 *
 * The debug messages that Alberto and Jerry need for testbed/cricket debugging
 * have been added.
 *
 * Revision 1.1  2000/03/08 01:31:40  jelson
 * RPC Software Reorganization
 *
 * Revision 1.7  1999/10/04 20:10:54  jelson
 * Cleanly split apart the library stub and the radiometrix radio daemon.
 * Client apps no longer have the capability of launching an rpcd on their
 * own, and, accordingly, don't need to carry around all that code with them.
 *
 * Revision 1.6  1999/10/03 01:05:42  jelson
 * SOme final bug fixes (WHOO HOO!!) and now everything seems to work, as a
 * first pass anyway.  The apitest/radiotest.c is a sample program that shows
 * how to use the interface.  It works on the real radios.  The corresponding
 * 'apitest.c' shows how to use Amit's server.
 *
 * Revision 1.5  1999/10/02 21:55:29  jelson
 * added time to debugging output
 *
 * Revision 1.4  1999/10/02 21:43:09  jelson
 * A few bug fixes in memory accessing and bounds checking
 *
 * Revision 1.3  1999/10/02 10:25:48  jelson
 * Major revision - the bulk of the code that implements the daemon launch,
 * client connect to daemon, and fragmentation code has been committed.
 * Code should be complete except for the last tiny layer of glue that passes
 * the reassmbled fragments back to localhost 2000.
 *
 * Revision 1.2  1999/09/28 07:20:07  intanago
 * Add the following lines due to SunOS's netinet/in.h doesn't define it.
 *
 * #ifndef INADDR_NONE
 * #define INADDR_NONE ((u_long) 0xffffffff)
 * #endif
 *
 * Revision 1.1  1999/09/28 04:32:02  jelson
 * Added code so that a radiometrix library client connects to the RPC
 * controller daemon using TCP, first creating the daemon if necessary.
 *
 * Added new test program 'radiotest' to test this.
 *
 *
 */

#define __UTIL_C__

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <fcntl.h>
#include <signal.h>
#include <ctype.h>
#include <limits.h>
#include <sys/time.h>


#include "rpcd_api.h"
#include "rpcd.h"


static int thiszone;
static char *debug_prefix = NULL;
int radiometrix_debug_level = DEFAULT_RDEBUG_LEVEL;

int32_t gmt2local(time_t t);

/* Set up what we need to print debugging messages. */
void rinit_debug(char *name)
{
  debug_prefix = MALLOC(char, strlen(name) + 16);
  sprintf(debug_prefix, "%s[%d]: ", name, (int) getpid());
  thiszone = gmt2local(0);
}
  

/* Print a debug message */
void rdebug_real(char *fmt, ...)
{
  va_list ap;
  char message[1024], time[64];
  struct timeval tv;
  int s;

  if (debug_prefix == NULL) {
    fprintf(stderr, "attempting to use debug function before init... sorry\n");
    return;
  }

  /* resolve var-arg buffer */
  va_start(ap, fmt);
  vsprintf(message, fmt, ap);

  /* get the time */
  gettimeofday(&tv, NULL);
  s = (tv.tv_sec + thiszone) % 86400;
  sprintf(time, "%02d:%02d:%02d.%06u ", s / 3600, (s % 3600) / 60, s % 60,
          /*      (uint32_t)tv.tv_usec); */
          (int)tv.tv_usec);

  /* put it all together */
  fprintf(stderr, "%s%s%s\n", time, debug_prefix, message);
  (void) fflush(stderr);

  va_end(ap);
}




void rdebug_print_message(int level, char *prefix, caddr_t message, int length)
{
  char buf[DEBUG_BUF_SIZE], c;
  int i;

  if (level > radiometrix_debug_level)
    return;

  strcpy(buf, prefix);
  strcat(buf, "'");

  for (i = strlen(buf); length && i < DEBUG_BUF_SIZE-100; length--, i++) {
    c = *((char *)message++);

    if (isprint(c))
      buf[i] = c;
    else
      buf[i] = '.';
  }
  buf[i] = '\0';

  strcat(buf, "'");
  rdebug_real(buf);
}

