/*
 * AFA fragmentation and reassembly library
 *
 * Jeremy Elson, USC/ISI
 *
 * For details on the theory of AFA fragmentation please see
 * http://www.circlemud.org/~jelson/writings/afa
 *
 * $Id: afa_fragmentation.c,v 1.10 2001/04/04 21:43:34 jelson Exp $
 */

/* KERNEL VERSION OF THE LIBRARY */
#ifdef __KERNEL__
#include <linux/kernel.h>
#include <linux/malloc.h>
#include <linux/random.h>
#include <linux/time.h>
#include "krpc.h"

#define MALLOC(type, length) kmalloc(length, GFP_ATOMIC)
#define FREE(x) kfree(x)

/* USER-LAND VERSION OF THE LIBRARY */
#else
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>
#include "rpcd.h"

/* MALLOC defined by rpcd.h */
#define FREE(x) free(x)
#endif /* __KERNEL__ */


/* our own includes */
#define __AFA_FRAGMENTATION_C__
#include "afa_fragmentation.h"


/* state about incoming packets */
rpc_packet_t packets[MAX_PACKETS];

#ifdef __KERNEL__

/* rand-like interface to the kernel's random number generation function */
static int rand()
{
  int n;

  get_random_bytes(&n, sizeof(n));
  return n;
}

#endif


void fragmentation_init()
{
  int i;

  for (i = 0; i < MAX_PACKETS; i++)
    packets[i].data = NULL;
}

void fragmentation_shutdown()
{
  int i;

  for (i = 0; i < MAX_PACKETS; i++)
    if (packets[i].data)
      FREE(packets[i].data);
}


/* Given a message to send, we fragment it into 24-byte chunks and
 * send each to the radio using our low-level radiometrix control
 * library. */
void fragment_message(caddr_t message, int length)
{
  frag_intro_t frag_intro;
  frag_data_t frag_data;
  int bytes_remaining;
  int curr_byte;
  caddr_t next_chunk;

  /* make sure we're not sending a message of length 0 */
  if (length <= 0 || length > MAX_PACKET_SIZE) {
    RDEBUG(1, "illegal fragmented-message size %d", length);
    return;
  }    

  /* first, send the frag intro */
  frag_intro.hdr.packet_type = FRAG_INTRO;
  frag_intro.hdr.frag_id = rand() % MAX_PACKETS;
  frag_intro.hdr.byte_id = length;
  frag_intro.checksum = calc_checksum(message, length);

  RDEBUG(10, "SENDING: queueing fragments, id=%d, len=%d, csum=%d, data=0x%s",
         frag_intro.hdr.frag_id, frag_intro.hdr.byte_id, frag_intro.checksum,
         asc2hex(message, length));

  RDEBUG(15, "SENDING: queued frag %d intro, len=%d, csum=%d",
         frag_intro.hdr.frag_id, frag_intro.hdr.byte_id, frag_intro.checksum);

  /* send the fragment out, using whatever method is provided by the
   * code linked with this file */
  send_fragment_upcall((caddr_t) &frag_intro, sizeof(frag_intro));

  /* now send each data packet until we get to the end */
  frag_data.hdr.packet_type = FRAG_DATA;
  frag_data.hdr.frag_id = frag_intro.hdr.frag_id;
  bytes_remaining = length;
  curr_byte = 0;
  next_chunk = message;

  while (bytes_remaining > 0) {
    int bytes_to_send;

    /* how much goes in this packet? */
    if (bytes_remaining <= FRAG_DATA_LEN)
      bytes_to_send = bytes_remaining;
    else
      bytes_to_send = FRAG_DATA_LEN;

    /* copy data in */
    frag_data.hdr.byte_id = curr_byte;
    memcpy(frag_data.data, next_chunk, bytes_to_send);

    RDEBUG(15, "SENDING: queued frag %d data [%d-%d]", frag_data.hdr.frag_id,
           frag_data.hdr.byte_id, bytes_to_send - 1 + frag_data.hdr.byte_id);

    send_fragment_upcall((caddr_t) &frag_data,
                         sizeof(frag_common_header_t) + bytes_to_send);

    /* remember what we sent */
    bytes_remaining -= bytes_to_send;
    curr_byte += bytes_to_send;
    next_chunk += bytes_to_send;
  }
  
}


/* We have received an incoming fragment.  Reassmble it; and, if it
 * completes a message, pass the message up to the next higher
 * layer. */
void process_incoming_fragment(caddr_t message, int length_with_header,
                               struct timeval packet_time)
{
  frag_common_header_t *frag_common = (frag_common_header_t *) message;
  int i, frag_id = frag_common->frag_id;

  if (frag_id < 0 || frag_id >= MAX_PACKETS) {
    RDEBUG(1, "out-of-range frag id %d is very confusing to me", frag_id);
    return;
  }

  switch(frag_common->packet_type) {

  /* this is a fragment INTRO - set up state required to accept the
   * rest of this data packet */
  case FRAG_INTRO: {
    frag_intro_t *frag_intro = (frag_intro_t *) message;
    int total_length = frag_intro->hdr.byte_id;

    /* make sure we're not allocing an insane amt of memory */
    if (total_length <= 0 || total_length > MAX_PACKET_SIZE) {
      RDEBUG(1, "received intro with illegal message size %d", total_length);
      return;
    }    

    RDEBUG(15, "RECEIVED: frag %d intro, len=%d, csum=%d", frag_id,
           total_length, frag_intro->checksum);

    if (packets[frag_id].data != NULL) {
      /* must be the remnant of an old, unfinished fragment... how sad */
      FREE(packets[frag_id].data);
      packets[frag_id].data = NULL; /* i'm paranoid :-) */
    }

    packets[frag_id].checksum = frag_intro->checksum;
    packets[frag_id].data = (caddr_t) MALLOC(char, total_length);
    packets[frag_id].length = total_length;
    packets[frag_id].time = packet_time;

    /* fill with "random" data to make sure the checksum fails until
     * the real packet arrives. */
    for (i = 0; i < total_length; i++)
      packets[frag_id].data[i] = (0x4a ^ i) & 0xff;

    break;
  }


  /* This is fragment DATA - put the data in its correct place, check
   * checksum, and if it passes, pass the packet up to the next layer */
  case FRAG_DATA: {
    frag_data_t *frag_data = (frag_data_t *) message;

    int start = frag_data->hdr.byte_id;
    int data_length = length_with_header - sizeof(frag_common_header_t);

    RDEBUG(15, "RECEIVED: frag %d data [%d-%d]", frag_data->hdr.frag_id,
            frag_data->hdr.byte_id, frag_data->hdr.byte_id + data_length - 1);

    /* is anything screwed up? */
    if (start < 0 || data_length <= 0) {
      RDEBUG(1, "RECEIVED frag with screwed up start/length");
      return;
    }

    /* did we miss the intro? */
    if (packets[frag_id].data == NULL) {
      RDEBUG(5, "got fragment %d with no state on record", frag_id);
      return;
    }

    /* did we miss the intro, AND we already have state from a
     * previous incomplete fragment with the same ID?  (whoa, unlikely
     * indeed).  In any case, just free state and forget it. */
    if (start + data_length > packets[frag_id].length) {
      RDEBUG(1, "warning - got fragment past end of allocated mem for packet");
      FREE(packets[frag_id].data);
      packets[frag_id].data = NULL;
      return;
    }

    /* okay... I think everything is safe now.  copy. */
    memcpy(packets[frag_id].data + start, frag_data->data, data_length);

    /* now check if the checksum matches... if so, THUNDERBIRDS ARE GO! */
    if (calc_checksum(packets[frag_id].data, packets[frag_id].length) ==
        packets[frag_id].checksum) {
      RDEBUG(10, "RECEIVED: reassembled fragments, id=%d, len=%d, csum=%d, data=0x%s",
             frag_id, packets[frag_id].length, packets[frag_id].checksum,
             asc2hex(packets[frag_id].data, packets[frag_id].length));
      deliver_packet_upcall(packets[frag_id].data, packets[frag_id].length,
                            packets[frag_id].time);
      FREE(packets[frag_id].data);
      packets[frag_id].data = NULL;
    } else {
      RDEBUG(30, "calc (%d) != intro (%d)", 
             calc_checksum(packets[frag_id].data, packets[frag_id].length),
             packets[frag_id].checksum);
    }

    break;
  }

  default:
    RDEBUG(1, "process_incoming_fragment: unknown frag type %d",
           frag_common->packet_type);
    break;

  }

}



u_int32_t calc_checksum(caddr_t message, int length)
{
  u_int32_t sum = 0;
  int i;

  /* stupid checksum function that i just made up - we'll replace it
   * with something real later */
  for (i = 0; i < length; i++)
    sum += (message[i] + i) * i;

  return sum;
}


#define MAXPACKET 400 /* max packet length that we can print */

char result[MAXPACKET*2];
char *asc2hex(caddr_t buffer, int length)
{
  int i;

  if (length < 0) {
    RDEBUG(0, "danger will robinson - length < 0 in asc2hex()!!!!!!!!\n");
    return "<DANGER>";
  }

  if (length >= MAXPACKET) {
    RDEBUG(1,"warning: insufficient buffer space in asc2hex for %d-byte data;"
             " data output is truncated", length);
    length = MAXPACKET-1;
  }

  for (i = 0; i < length; i++) {
    sprintf(result+(2*i), "%x%x", (buffer[i] & 0xf0) >> 4,
            (buffer[i] & 0x0f));
  }

  return result;
}
#undef MAXPACKET
