/* #define WANT_DEBUG */ /* uncomment this line for debug output */

#ifndef WANT_DEBUG
#define DEBUG(x)
#define DEBUG2(x, y)
#endif

#ifdef __KERNEL__
#include <linux/delay.h>
#include <linux/types.h>
#define usleep udelay

#ifdef WANT_DEBUG
#include <linux/kernel.h>
#define DEBUG(x) printk(x)
#define DEBUG2(x, y) printk(x, y)
#endif

#else /* userspace */

#ifdef WANT_DEBUG
#include <stdio.h>
#define DEBUG(x) fprintf(stderr, x)
#define DEBUG2(x, y) fprintf(stderr, x, y)
#endif

#include <unistd.h>
#include <sys/types.h>
#endif

#define __RPC_LOWLEVEL_C__

#include "parapin.h"
#include "rpc_lowlevel.h"

/* Mappings of RPC pins to PC parallel port pins as suggested by the
 * Radiometrix Application Note */
#define D0  LP_PIN02
#define D1  LP_PIN03
#define D2  LP_PIN04
#define D3  LP_PIN05

#define TXR LP_PIN16
#define TXA LP_PIN12
#define RXR LP_PIN13
#define RXA LP_PIN14

#define RST LP_PIN17

static int count;

#define WAIT_WHILE(condition) \
  do { count=0; \
       while ((condition) && (count < RPC_TIMEOUT_INTERVAL)) count++; \
       if (condition)  goto timeout; \
   } while(0)

/**************************************************************************/


/* Initializes pins to a sane state but doesn't try to send anything.
 * Should only be called once when the app starts.  Always
 * succeeds. */
static void init_parallel_pins()
{
  /* Put pins in a sane state */
  pin_input_mode(LP_DATA_PINS);
  pin_input_mode(LP_SWITCHABLE_PINS);
  pin_output_mode(RXA | TXR | RST);
  clear_pin(RXA | TXR | RST);
}


/* Receive a single nibble from the radiometrix */
static int receive_nibble()
{
  int pin_state, retval;

  DEBUG("......RECEIVING NIBBLE - start\n");

  /* wait for data to be ready */
  WAIT_WHILE(pin_is_set(RXR));

  /* tell the RPC we're ready to receive */
  clear_pin(RXA);

  /* wait for the RPC to assert its data lines */
  WAIT_WHILE(!pin_is_set(RXR));

  /* read the data */
  pin_state = pin_is_set(D0 | D1 | D2 | D3);
  retval = 0;
  retval |= (pin_state & D0) ? (1 << 0) : 0;
  retval |= (pin_state & D1) ? (1 << 1) : 0;
  retval |= (pin_state & D2) ? (1 << 2) : 0;
  retval |= (pin_state & D3) ? (1 << 3) : 0;

  /* tell the RPC we've got this nibble */
  set_pin(RXA);

  DEBUG("......RECEIVING NIBBLE - end\n");

  /* return value to caller */
  return retval;

timeout:
  return RPC_TIMEOUT;
}



/* Receive a single byte from the radiometrix */
static int receive_byte()
{
  int retval;
  int nibble;

  DEBUG("...RECEIVING BYTE - start\n");

  /* get least significant nibble first */
  if ((nibble = receive_nibble()) == RPC_TIMEOUT)
    return RPC_TIMEOUT;

  retval = nibble;

  /* then get most significant nibble */
  if ((nibble = receive_nibble()) == RPC_TIMEOUT)
    return RPC_TIMEOUT;

  retval |= (nibble << 4);

  DEBUG("...RECEIVING BYTE - end\n");
  return retval;
}


/* Poll the radiometrix to determine if a message is waiting.  If so,
 * get it, and call the user-supplied callback if any.  Return value
 * of 1 indicates that a message was waiting and the upcall has been
 * called; 0 means no message was waiting.  RPC_TIMEOUT means an error
 * occurred (usually a timeout) and the RPC should be reset.  */
int perform_receive_message()
{
  char buf[30];
  int header, new_byte;
  int length;
  int i;

  /* if the RPC is not trying to upload something to us, do nothing */
  if (pin_is_set(RXR))
    return 0;

  DEBUG("RECEIVING MESSAGE - start\n");

  /* Get the first byte - this is the header */
  if ((header = receive_byte()) == RPC_TIMEOUT)
    return RPC_TIMEOUT;

  /* The low 5 bits of the header indicate the length of the message
     including the header byte.  Now we determine the length of the
     data part. */
  length = (header & 0x1f) - 1; /* 0x1f = binary 00011111 */

  /* This should never happen... */
  if (length < 0 || length > 27) {
    DEBUG2("WARNING: received strange %d-length message?\n", length);
    return RPC_TIMEOUT;
  }

  DEBUG2("...GOT header.  %d byte message following\n", length);

  /* Now that we know how long the message is, receive as many bytes
   * as required. */
  for (i = 0; i < length; i++) {
    if ((new_byte = receive_byte()) == RPC_TIMEOUT)
      return RPC_TIMEOUT;
    buf[i] = new_byte;
  }

  DEBUG("RECEIVING MESSAGE - end\n");

  /* the code linked with this library must provide this function */
  rpc_incoming_upcall(buf, length);

  return 1;
}



/***************************** Sending **********************************/


static int send_nibble(char data)
{
  DEBUG("......SEND NIBBLE - start\n");

  /* tell the RPC we want to upload a message to it */
  clear_pin(TXR);

  /* now, wait until the RPC pulls TXA low.  If RXR goes low instead,
   * we must process the incoming message first.  According to the
   * radiometrix data sheet: "The RPC will ignore a TX request from
   * the Host while it is receiving a packet from the radio.  [...] If
   * a valid packet is received this must be uploaded to the Host
   * before the RPC can respond to the Host's TX request.  Thus an RX
   * Request will be signalled to the Host and not the expected TX
   * Accept, and the Host must upload the incoming packet before the
   * TX packet can be downloaded.  The TX Request should be left
   * asserted (low) during the upload."  */
  do {
    WAIT_WHILE(pin_is_set(TXA) && pin_is_set(RXR));
    if (!pin_is_set(RXR))
      perform_receive_message();
  } while (pin_is_set(TXA));


  /* assert the data on D0-D4 */
  pin_output_mode(LP_DATA_PINS);
  change_pin(D0, (data & 0x01) ? LP_SET : LP_CLEAR);
  change_pin(D1, (data & 0x02) ? LP_SET : LP_CLEAR);
  change_pin(D2, (data & 0x04) ? LP_SET : LP_CLEAR);
  change_pin(D3, (data & 0x08) ? LP_SET : LP_CLEAR);

  /* tell the RPC the data is available */
  set_pin(TXR);

  /* wait for the data to be accepted */
  WAIT_WHILE(!pin_is_set(TXA));

  /* change drivers back to input mode */
  pin_input_mode(LP_DATA_PINS);

  DEBUG("......SEND NIBBLE - end\n");
  return 0;

timeout:
  return RPC_TIMEOUT;
}


static int send_byte(char data)
{
  DEBUG("...SEND BYTE - start\n");

  /* send least signficant nibble */
  if (send_nibble(data & 0x0f) == RPC_TIMEOUT)
    return RPC_TIMEOUT;

  /* send most significant nibble */
  if (send_nibble((data & 0xf0) >> 4) == RPC_TIMEOUT)
    return RPC_TIMEOUT;

  DEBUG("...SEND BYTE - end\n");
  return 0;
}


/* send a message out to the rpc.  returns the number of bytes written
 * if succesful, or RPC_TIMEOUT if there's an error (including a
 * timeout) */
int perform_send_message(caddr_t message, int length)
{
  char header;
  int remaining;

  /* trying to send too long of a message? */
  if (length > 27 || length <= 0)
    return -1;

  DEBUG("SEND MESSAGE - start\n");

  /* high 3 bits have special meaning.  low 5 bits are length
     including this header byte */
  header = length + 1;

  if (send_byte(header) == RPC_TIMEOUT)
    return RPC_TIMEOUT;

  for (remaining = length; remaining > 0; remaining--) 
    if (send_byte(*(message++)) == RPC_TIMEOUT)
      return RPC_TIMEOUT;

  DEBUG("SEND MESSAGE - end\n");

  return length;
}


/* these functions help split sending into two stages: first, telling
 * the RPC that we have something to send (by clearing TXR); then,
 * polling later to see if the RPC is ready for us to upload the
 * packet. */
void perform_tx_request()
{
  clear_pin(TXR);
}

/* is the RPC ready to receive?  1 if ready, 0 if not.  valid only if
 * perform_tx_request was previously called.  to maintain sanity,
 * perform_send_message or perform_send_control must follow
 * perform_tx_poll if it returns 1!  */
int perform_tx_poll()
{
  if (pin_is_set(TXA))
    return 0;
  else
    return 1;
}

/* call this if you need to cancel a TX request after
 * perform_tx_request and perform_tx_poll.  it works just by stupidly
 * re-sending a control byte that doesn't do anything.
 */
int perform_tx_cancel()
{
  return set_rpc_switches();
}


/* Sends control bytes to the RPC.  Returns 0 on success or
 * RPC_TIMEOUT on failure such as timeout. */
int perform_send_control(char address, char value)
{
  char header;

  /* set high 2 bits (1 << 7 == Packet Type; 1 << 6 == Write), and low
   * 6 bits of address */
  header = (address & 0x3f) | (1 << 7) | (1 << 6);

  if (send_byte(header) == RPC_TIMEOUT)
    return RPC_TIMEOUT;

  if (send_byte(value) == RPC_TIMEOUT)
    return RPC_TIMEOUT;

  return 0; /* success */
}



/************************** Other Utilities *****************************/

/*
 *  The command below activates "Random Delay after Channel Clear" TX
 *  mode, and turns off all other modes (power saving, host timeout,
 *  radio timeout, self test, etc.)
 *
 * This is a separate function so it can be called to cancel a send
 * (basically a meaningless control byte in case we want to cancel a
 * perform_tx_request() after it has been accepted)a
 */
static int set_rpc_switches()
{
  return perform_send_control(0, (1 << 1) | (1));
}


/* Reset the RPC.  Returns 0 on success, or RPC_TIMEOUT on timeout. */
int perform_reset_rpc()
{
  /* set pins to their state required on reset */
  init_parallel_pins();

  /* From RPC documentation:
   *
   * HOST DRIVEN RESET Minimum low time: 1.0 ms, after reset is
   * released (returned high). The host should allow a delay 1ms after
   * reset for the RPC to initialise itself During this delay the host
   * must hold TXR high (unless DIAGNOSTIC MODES are required) and RXR
   * signal should be ignored.
   */
  set_pin(RXA | TXR);
  clear_pin(RST);
  usleep(1000);
  set_pin(RST);
  usleep(1000);

  /* now set switches as part of initialization; can time out */
  return set_rpc_switches();
}

